/*-------------------------------------------------------------------*/
/* XMEANS.C     Karol Waga / Radu Mariescu                           */
/*                                                                   */
/* X-means implementation                                            */
/*                                                                   */
/* Naming conventions used in the code                               */
/*                                                                   */
/*    TS        training set (data objects)                          */
/*    CB        codebook (cluster representatives, centroids)        */
/*    P         partitioning (pointing from TS to CB)                */
/*                                                                   */
/*    p-prefix  pointer, e.g. pTS is pointer to the training set TS  */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName       "XMEANS"
#define VersionNumber  "Version 0.05"
#define LastUpdated    "16.01.2013"

/* ------------------------------------------------------------------- */
/*      Changelog:                                                     */
/*                                                                     */
/* RMI    0.02 Added additional stopping criteria  rep_max             */
/* RMI    0.03 Solved problem with dividing 1 vector to 2 clusters     */
/* RMI    0.04 Reorganising code                                       */
/*                added CalculateGlobalVectorToCentroidDistance()      */
/*                added CalculateLogLikelihood()                       */
/*                    rewrote BIC() & AIC()                            */
/*                changes in PerformXMeans(), GetSubspaceForCentroid() */
/* KW,RMI 0.05 Runs up to given number of clusters if min and max      */
/*             cluster numbers are equal            				   */
/* RMI,KW 0.06 Forcing a split even when AIC/BIC criterion isn't met   */
/*             IF we know number of clusters                           */  
/*                                                                     */
/* ------------------------------------------------------------------- */
/*-------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "cb.h"
#include "random.h"
#include "interfc.h"
#include "reporting.h"
#include "rls.h"
#define PI 3.14159265


/* =================== PROTOTYPES =================================== */

int PerformXMeans(TRAININGSET *pTS, CODEBOOK *pCB, PARTITIONING *pP,
          int minClus, int maxClus, int kRep, int splitCount, int criterion, int InitMethod, int quietLevel, int useInitial);
static void InitializeSolutions(TRAININGSET *pTS, CODEBOOK *pCB, 
    PARTITIONING *pP, CODEBOOK *pCBnew, PARTITIONING *pPnew, 
    CODEBOOK *pCBinit, PARTITIONING *pPinit, llong *distanceInit, 
    int clus, int initial);
static void GenerateSolution(TRAININGSET *pTS, CODEBOOK *pCBnew, 
    PARTITIONING *pPnew, CODEBOOK *pCBinit, PARTITIONING *pPinit, 
                 llong *distance, llong *distanceInit, 
                 int InitMethod, int initial);
static void KMeansIterate(TRAININGSET *pTS, CODEBOOK *pCBnew, 
    PARTITIONING *pPnew, llong *distance, int quietLevel, int i, 
    int *iter, double *time, double *error, int initial);
static void GetSubspaceForCentroid(TRAININGSET *tmpTS, PARTITIONING *Pnew, PARTITIONING *Pnew1, CODEBOOK *CBnew, CODEBOOK *CBnew1, TRAININGSET *pTS, int j, int i);
static long BIC(TRAININGSET *TS, CODEBOOK *CB, PARTITIONING *P);
static long AIC(TRAININGSET *TS, CODEBOOK *CB, PARTITIONING *P);
char* XMeansInfo(void);
static double CalculateGlobalVectorToCentroidDistance(TRAININGSET *TS, CODEBOOK *CB, PARTITIONING *P);
static double CalculateLogLikelihood(TRAININGSET *TS, CODEBOOK *CB, PARTITIONING *P, double variance);

/* =================== PUBLIC FUNCTIONS ============================= */

/* Gets training set pTS (and optionally initial codebook pCB or 
   partitioning pP) as a parameter, generates solution (codebook pCB + 
   partitioning pP) and returns 0 if clustering completed successfully. 
   N.B. Random number generator (in random.c) must be initialized! */

int PerformXMeans(TRAININGSET *pTS, CODEBOOK *pCB, PARTITIONING *pP,
          int minClus, int maxClus, int kRep, int splitCount, int criterion, int InitMethod,  
          int quietLevel, int useInitial)
{
  TRAININGSET    tmpTS;
  PARTITIONING  Pnew, Pinit, tempP, Pnew2, Pnew1;
  CODEBOOK      CBnew, CBinit, tempCB, CBnew2, CBnew1;
  llong         distance[BookSize(pTS)]; 
  llong         distanceInit[BookSize(pTS)]; 
  double        totalTime, error, currError;
  int           i, old_i=-1, i_rep=0, rep_max=1000, j, k, better, iter, 
                totalIter, numClus=0, first=1;
  int             new_score, old_score, total_new_score, total_old_score=0, Score=0;
  int			equalClus = 0;
  SetClock(&totalTime);
  totalIter = 0;
  currError = error = 0;

  if ((minClus < 1) || (BookSize(pTS) < minClus) || (maxClus < 1) || (BookSize(pTS) < maxClus) || (minClus > maxClus))
  {
    return 1;   /* clustering failed */
  }
  
  if (minClus==maxClus){
	minClus = 1;
	equalClus = 1;
  }
            
  PrintHeader(quietLevel);
  
  //starting to search from minimal range
  i=minClus;
    
  //initializing random codebook with minimal range number of clusters
  InitializeSolutions(pTS, pCB, pP, &CBnew, &Pnew, &CBinit, &Pinit, 
         distanceInit, i, useInitial);
  GenerateSolution(pTS, &CBnew, &Pnew, &CBinit, &Pinit, distance, 
         distanceInit, InitMethod, useInitial);
    
	
	
  //repeat until maximum range    
  //while(i<maxClus)// && i_rep<rep_max)
  while ((i<maxClus && equalClus==1)||(i<maxClus && i_rep<rep_max && equalClus==0))
  {
    better = iter = 0;
    totalIter=0;
    
    for(j = 0; j < (&Pnew)->PartitionCount; j++ )
    {
      GetSubspaceForCentroid(&tmpTS,&Pnew,&Pnew1,&CBnew,&CBnew1,pTS,j,i);
      
      //calculating score before split
      (criterion==1)?(old_score = BIC(&tmpTS,&CBnew1,&Pnew1)):(old_score = AIC(&tmpTS,&CBnew1,&Pnew1));
        
      //initializing the Subspace trainingset with 2 cluster centroids
      InitializeSolutions(&tmpTS, &tempCB, &tempP, &CBnew2, &Pnew2, &tempCB, &tempP, 
             distanceInit, splitCount, useInitial);
      if(BookSize(&tmpTS)==1)//if you can't split into two partitions, because there is only 1 vector 
        continue;
      GenerateSolution(&tmpTS, &CBnew2, &Pnew2, &tempCB, &tempP, distance, 
             distanceInit, InitMethod, useInitial);
      
      //run k-means with kRep iterations 
      KMeansIterate(&tmpTS, &CBnew2, &Pnew2, distance, quietLevel, kRep, &iter, 
             &totalTime, &error, useInitial);
      totalIter += iter;
        
      //calculating score after split
      (criterion==1)?(new_score=BIC(&tmpTS,&CBnew2,&Pnew2)):(new_score=AIC(&tmpTS,&CBnew2,&Pnew2));
        
      //if criterion gives the GO, move solution from Subspace to global Codebook
      if(old_score-new_score<0 || i==1 || i_rep==rep_max)
      {
	    i_rep=0;
        //keepSolution
        CopyVector(Vector(&CBnew2, 0), Vector(&CBnew, j), VectorSize(&CBnew2));
        for(k=0;k<splitCount-1;k++)
        {
          IncreaseCodebookSize(&CBnew,BookSize(&CBnew)+1);
          CBnew.Book[BookSize(&CBnew)-1].freq=0;
          CopyVector(Vector(&CBnew2, (k+1)), Vector(&CBnew, BookSize(&CBnew)-1), VectorSize(&CBnew2));
          i++;// increment number of centroids		  
		  if (i>=maxClus){		  
			break;
		  }
			
        }
      }
        
      //free unneeded objects from memory
      FreeCodebook(&tmpTS);
      FreeCodebook(&CBnew2);
      FreeCodebook(&tempCB);
      FreePartitioning(&Pnew2);
      FreePartitioning(&tempP);
      
      (i==old_i)?(i_rep++):(i_rep=0);
      old_i=i;
      
      //no point to split anymore, boundary limitation OR no improvement for long time
      if(i>=maxClus)// || i_rep>=rep_max)
        break;
    } 
    
    //create partitioning for new centroid locations
    CreateNewPartitioning(&Pnew, pTS, i);
    GenerateOptimalPartitioningGeneral(pTS, &CBnew, &Pnew, MSE);
    //run k-means until stabilized
    KMeansIterate(pTS, &CBnew, &Pnew, distance, quietLevel, -1, &iter, 
            &totalTime, &error, useInitial);
            
    totalIter += iter;
        
    //if score better than any previous
    (criterion==1)?(total_new_score=BIC(pTS,&CBnew,&Pnew)):(total_new_score=AIC(pTS,&CBnew,&Pnew));
    
    if (total_old_score-total_new_score<0 || first==1 || ((i==(maxClus)) && (equalClus==1))) 
    {
      first=0;
      total_old_score=total_new_score;
      CreateNewCodebook(pCB, i, pTS);
      CopyCodebook(&CBnew, pCB);
      CreateNewPartitioning(pP, pTS, i);
      CopyPartitioning(&Pnew, pP);
      currError = error;
      better = 1;
      numClus=i;
      Score=-total_new_score;
    }
    PrintXM(quietLevel, i, totalIter, error, GetClock(totalTime), better);
  }

  PrintFooterXM(quietLevel, currError, numClus, Score, criterion, GetClock(totalTime), totalIter);

  FreeCodebook(&CBnew);
  FreePartitioning(&Pnew);
  FreeCodebook(&CBinit);
  FreePartitioning(&Pinit);

  return 0;
}  /* PerformXMeans() */

/* ------------------------------------------------------------------ */

char* XMeansInfo(void)
{
  char* p;
  int len;
  
  len = strlen(ProgName)+strlen(VersionNumber)+strlen(LastUpdated)+4;  
  p   = (char*) malloc(len*sizeof(char));
  
  if (!p) 
  {
    ErrorMessage("ERROR: Allocating memory failed!\n");
    ExitProcessing(FATAL_ERROR);
  }
  
  sprintf(p, "%s\t%s\t%s", ProgName, VersionNumber, LastUpdated);
  
  return p;
}  /* XMeansInfo() */


/* ====================== PRIVATE FUNCTIONS ========================= */

static void InitializeSolutions(TRAININGSET *pTS, CODEBOOK *pCB, 
PARTITIONING *pP, CODEBOOK *pCBnew, PARTITIONING *pPnew, CODEBOOK *pCBinit, 
PARTITIONING *pPinit, llong *distanceInit, int clus, int initial)
{
  CreateNewCodebook(pCBnew, clus, pTS);
  CreateNewPartitioning(pPnew, pTS, clus);
  CreateNewCodebook(pCBinit, clus, pTS);
  CreateNewPartitioning(pPinit, pTS, clus);
  
  if (initial) 
  {  
    if (initial == 1)
    {
      GenerateOptimalPartitioningGeneral(pTS, pCB, pP, MSE);
    } 
    else if (initial == 2)
    {
      GenerateOptimalCodebookGeneral(pTS, pCB, pP, MSE);
    } 
      
    CopyCodebook(pCB, pCBinit);
    CopyPartitioning(pP, pPinit);
    CalculateDistances(pTS, pCBinit, pPinit, distanceInit);
  }
}  /* InitializeSolutions() */

/* ------------------------------------------------------------------ */

static void GenerateSolution(TRAININGSET *pTS, CODEBOOK *pCBnew, 
                 PARTITIONING *pPnew, CODEBOOK *pCBinit, 
                 PARTITIONING *pPinit, llong *distance, 
                 llong *distanceInit, int InitMethod, int initial) 
{
  int i;
  if (initial)
  {
    CopyCodebook(pCBinit, pCBnew);
    CopyPartitioning(pPinit, pPnew);
    for (i = 0; i < BookSize(pTS); i++)
    {
      distance[i] = distanceInit[i];      
    }
  }
  else
  {
    if (InitMethod == 0) 
      SelectRandomRepresentatives(pTS, pCBnew);
    else if (InitMethod == 1)
      SelectRandomWeightedRepresentatives(pTS, pCBnew);
    else if (InitMethod == 2)
      SelectRandomRepresentativesbyMarko(pTS, pCBnew);
    else
      SelectRandomRepresentatives2(pTS, pCBnew);

    GenerateOptimalPartitioningGeneral(pTS, pCBnew, pPnew, MSE);    
    CalculateDistances(pTS, pCBnew, pPnew, distance);
  }
}  /* GenerateSolution() */

/* ------------------------------------------------------------------ */

static void GetSubspaceForCentroid(TRAININGSET *tmpTS, PARTITIONING *Pnew, PARTITIONING *Pnew1, CODEBOOK *CBnew, CODEBOOK *CBnew1, TRAININGSET *pTS, int j, int i)
{
  CreateNewTrainingSet(tmpTS,
                       UniqueVectors(Pnew,j),
                       pTS->BlockSizeX,
                       pTS->BlockSizeY,
                       pTS->BytesPerElement,
                       pTS->MinValue,
                       pTS->MaxValue,
                       pTS->Preprocessing,
                       pTS->GenerationMethod);
      
  int m=0, k;
  for(k=0;k<BookSize(pTS);k++){
    if(Map(Pnew,k)==j){
      CopyNode(&Node(pTS,k), &Node(tmpTS,m), VectorSize(pTS));
      VectorFreq(tmpTS, m) = 1;
      m++;
    }
  }
        
  CreateNewCodebook(CBnew1, 1, tmpTS);
  CopyVector(Vector(CBnew, j), Vector(CBnew1, 0), VectorSize(CBnew));
  CBnew1->Book[0].freq=0;
     
  CreateNewPartitioning(Pnew1, tmpTS, i);
  GenerateOptimalPartitioningGeneral(tmpTS, CBnew1, Pnew1, MSE);
} /* GetSubspaceForCentroid() */

/* ------------------------------------------------------------------ */

static double CalculateGlobalVectorToCentroidDistance(TRAININGSET *TS, CODEBOOK *CB, PARTITIONING *P)
{
  int j,k;
  double dist=0;
  for(j = 0; j < P->PartitionCount; j++ )
  {
    for(k=0;k<BookSize(TS);k++)
    {
      if(Map(P,k)==j)
      {
        dist+=VectorDistance(Vector(CB,Map(P,k)),
                                  Vector(TS,k),
                                  VectorSize(TS),
                                  MAXLLONG,
                                  EUCLIDEANSQ)   * VectorFreq(TS,k);
      }
    }
  }
  return dist;
}

/* ------------------------------------------------------------------ */

static double CalculateLogLikelihood(TRAININGSET *TS, CODEBOOK *CB, PARTITIONING *P, double variance)
{
  double loglik=0;
  int j,k;
  for(j = 0; j < P->PartitionCount; j++ )
  {
    int owned=0;
    double dist_curr=0;
    for(k=0;k<BookSize(TS);k++){
      if(Map(P,k)==j){
        owned++;
        dist_curr+=VectorDistance(Vector(CB,Map(P,k)),
                                  Vector(TS,k),
                                  VectorSize(TS),
                                  MAXLLONG,
                                  EUCLIDEANSQ)   * VectorFreq(TS,k);
      }
    }
    if(owned>1){//log(1)=0 so we exclude
      double loglik_curr = owned*log(owned)
        - owned*log(BookSize(TS))
        - owned/2.0*log(2*PI)
        - owned*VectorSize(TS)/2.0*log(variance)
        - dist_curr/(2*variance);
      loglik += loglik_curr;
    }
  }
  return loglik;
}

/* ------------------------------------------------------------------ */

static long BIC(TRAININGSET *TS, CODEBOOK *CB, PARTITIONING *P)
{
  int num_parameters = (BookSize(CB) - 1) +
            BookSize(CB) * VectorSize(TS) +    
            BookSize(CB);    
  double log_owned = log((double) BookSize(TS));
  
  double dist=CalculateGlobalVectorToCentroidDistance(TS,CB,P);
  double variance=dist/(1.0*BookSize(TS) - BookSize(CB));
  
  double loglik=CalculateLogLikelihood(TS,CB,P,variance);
  
  return loglik - (num_parameters/2.0)*log_owned;
} /* BIC() */

/* ------------------------------------------------------------------ */

static long AIC(TRAININGSET *TS, CODEBOOK *CB, PARTITIONING *P)
{
  int num_parameters = (BookSize(CB) - 1) +
            BookSize(CB) * VectorSize(TS) +    
            BookSize(CB);    
  
  double dist=CalculateGlobalVectorToCentroidDistance(TS,CB,P);
  double variance=dist/(1.0*BookSize(TS) - BookSize(CB));
  
  double loglik=CalculateLogLikelihood(TS,CB,P,variance);
  
  return loglik - (num_parameters/2.0)*2;
} /* AIC() */

/* ------------------------------------------------------------------ */

static void KMeansIterate(TRAININGSET *pTS, CODEBOOK *pCBnew, 
PARTITIONING *pPnew, llong *distance, int quietLevel, int i, int *iter, 
double *time, double *error, int initial)
{
  int       active[BookSize(pCBnew)];
  int       activeCount;
  double    oldError;

  *iter  = 0;
  *error = AverageErrorForSolution(pTS, pCBnew, pPnew, MSE);
  PrintIterationKM(quietLevel, i, *iter, *error, GetClock(*time));

  /* K-means iterations */
  do 
  {
    (*iter)++;
    oldError = *error;

    OptimalRepresentatives(pPnew, pTS, pCBnew, active, &activeCount);
    OptimalPartition(pCBnew, pTS, pPnew, active, activeCount, distance);

    *error = AverageErrorForSolution(pTS, pCBnew, pPnew, MSE);
    PrintIterationKM(quietLevel, i, *iter, *error, GetClock(*time));
  }
  while (*error < oldError && (*iter)!=i);  /* until no improvement */
}  /* KMeansIterate() */

